--[[
	Module with TerraME Utility functions for cell spaces
	These function can be reused - they have functions that can be used in different TerraME models
]]
tmepath = os.getenv("TME_PATH_1_3_0")
package.path = package.path .. ";"..tmepath.."/contrib/terrameaddons/?.lua"
local rand = require ("random")

local cells ={}
cells.version="1.0"

--[[ 
	Manhattan distance
]]
function cells.blockDist (c1, c2)
	return math.abs(c1.x - c2.x) + math.abs (c1.y - c2.y)
end

--[[  
	Get an random cell in the cell space that is not occupied by any other agent
]]
function cells.findEmpty (cs)
	found = false
	while not found do
		i = rand:integer (1, cs:size())     -- get a random cell index
		c = cs:getCells()[i]                -- get a random cell
		if (#(c:getAgents()) == 0 ) then    -- is the cell empty? 
			found = true
		end
	end
return c
end

--[[
	Create a rook neighborhood of a given length
	The cellspace is a torus, so side cells are connected.
]]

-- create MxN Neighborhood 

function cells.createRookNeighborhood(cs, len, dim)
	-- len : size of neighborhood in cells (von neumann has a size of 1)
	-- dim : dimension of the cell space
	assert (len >= 1,  "createRookNeighborhood: length is not a positive integer")
	assert (2*len < dim, "createRookNeighborhood: size is greater then cellspace dimension")
    
    cs:createNeighborhood { 
        strategy = "mxn",  
        m = len, 
        n = len,
        name = string.format ("neigh%d", len),
        weight = function (c1, c2)
                    return 1
                 end,
        filter = function (c1, c2)
                    return (c1.x == c2.x) or (c1.y == c2.y ) 
                 end
        } 
end

-- find path between cells of the same cell space
function cells.findPath(c1, c2)
	local path = {}
    local xp = c1.x
    local yp = c1.y
    local distX = c2.x - c1.x
    local distY = c2.y - c1.y
    if (distX > 0) then stepX = 1 else stepX = -1 end
	if (distY > 0) then stepY = 1 else stepY = -1 end
                
    repeat
        if (distX ~= 0 ) then
            xp = xp + stepX
            distX = distX - stepX
			table.insert (path, Coord {x = xp, y = yp})
        end
        if (distY ~= 0 ) then
            yp = yp + stepY
            distY = distY - stepY
			table.insert (path, Coord {x = xp, y = yp})
        end
    until (distX == 0 ) and (distY == 0 )       
return path
end

--read a cellspace from a file 

function cells.fromTxt (argv)
 	-- create a cellular space from a text file
 	-- xdim: horizontal dimension of the cell space
 	-- ydim: vertical dimension of the cell space
 	-- attr_name: name of the attribute to be read from text file
 	-- txtfile: text file - MUST have EXACTLY ydim lines
 	-- charsize: size of each character to be read
 	-- whitespaces: number of whitespaces between characters

	local cs = CellularSpace {
		xdim = argv.xdim,
		ydim = argv.ydim
	}
	local i = 0
	for line in io.lines(argv.txtfile) do
		assert (#line == (argv.xdim*(argv.charsize + argv.whitespaces) - 1), "csFromTxt: wrong input")
		j  = 0 		
		ic = 1
		repeat
			local p = Coord { x = j, y = i } -- TerraME arrays start at 0, Lua's at 1 
			cs:getCell(p)[argv.attrname] = tonumber (string.sub (line, ic, ic + argv.charsize - 1))
			j = j + 1
			ic = ic + argv.charsize + argv.whitespaces
		until ic > #line
		i = i + 1
	end
return cs
end

function cells.northSouthTrajectories (cs)
	local north = Trajectory { 
		target = cs,
		select = function (cell)
			return cell.y < (DIM_SPACE/2)
		end
	}
	local south = Trajectory { 
		target = cs,
		select = function (cell)
			return cell.y >= (DIM_SPACE/2)
		end
	}
return north, south
end

function cells.initCellSpace (cs, attr, value)
    forEachCell (cs, function (cell) cell[attr] = value end )
end

function cells.countNeighbors(cell, attr, value)
    local count = 0 
    forEachNeighbor (cell, 
                        function (cell, neigh) 
                            if (neigh[attr] == value) then
                                count = count + 1
                            end
                        end
                    )
    return count
end

--[[ 
function to count the past neighbors of a cell that are in a given state
	always reads from the past copy of the CA
]]
function cells.countPastNeighbors(cell, attr, val)
	count = 0
	forEachNeighbor(cell, function(cell, neigh)
		if neigh.past[attr] == val then
			count = count + 1
		end
	end)
	return count
end

return cells
