-- Simple statistical functions              --
----------------------------------------------
local ms = require ("moses")

-- Table to hold statistical functions
local stats={}

-- Get the mean value of a table of values
function stats.mean( vs )
  local sum   = 0
  local count = 0

  for k,v in pairs(vs) do
    if type(v) == 'number' then
      sum = sum + v
      count = count + 1
    end
  end
  return (sum / count)
end

-- Get the mode of a table.  Returns a table of values.
-- Works on anything (not just numbers).
function stats.mode(vs)
    local counts={}

    for k,v in pairs (vs) do
        if counts[v] == nil then
            counts[v] = 1
        else
            counts[v] = counts[v] + 1
        end
    end

    local biggestCount = 0

    for k,v  in pairs (counts) do
        if v > biggestCount then
            biggestCount = v
        end
    end

    local temp={}

    for k,v in pairs( counts ) do
        if v == biggestCount then
            table.insert( temp, k )
        end
    end

    return temp
end

-- Get the median of a table.
function stats.median (vs)
    table.sort (vs)
    mid = #vs/2
    -- If we have an even number of table elements or odd.
    if #vs % 2 == 0 then
    -- return mean value of middle two elements
        return (vs[mid] + vs[mid + 1])/2
    else
        -- return middle element
        return vs[math.ceil(mid)]
    end
end
    

-- Get the standard deviation of a table
function stats.stdev( t )
  local m
  local vm
  local sum = 0
  local count = 0
  local result

  m = stats.mean( t )

  for k,v in pairs(t) do
    if type(v) == 'number' then
      vm = v - m
      sum = sum + (vm * vm)
      count = count + 1
    end
  end

  result = math.sqrt(sum / (count-1))

  return result
end

-- Get the max and min for a table
function stats.minmax( t )
  local max = -math.huge
  local min = math.huge

  for k,v in pairs( t ) do
    if type(v) == 'number' then
      max = math.max( max, v )
      min = math.min( min, v )
    end
  end

  return min, max
end

-- Get the max  for a table
function stats.max( t )
  local max = -math.huge

  for k,v in pairs( t ) do
    if type(v) == 'number' then
      max = math.max( max, v )
    end
  end

  return max
end

function stats.min( t )
  local min = math.huge

  for k,v in pairs( t ) do
    if type(v) == 'number' then
      min = math.min( min, v )
    end
  end

  return min
end

--[[
  Calculate a cumulative distribution function 
]]
function stats.distrib (t)
    assert (#t >= 1, "distrib: number of entries should be positive")
    table.sort (t) 

    min, max = stats.minmax (t)
    assert (min >= 1, "distribution: minimum value should be greater than 1")

    local d = {}
    for e = 1, max do
        d[e] = ms.count (t, e) 
    end

return d 
end 


--[[
  Calculate a histogram
]]

function stats.histogram (t, steps)
    assert (steps > 0, "histogram: steps must be a positive integer")
   
    min, max = tablib.minmaxKeys (t)
    assert (min >= 1,   "histogram: only works for positive integers")

    step = math.ceil ( max /steps )
    
    local hist = {}
    for i = 1, steps do
        local bin = {}
        bin.min = (i - 1)*step + 1
        bin.max = i*step 
        bin.num = 0 
        for k, v in pairs (t) do
            if (k >= bin.min) and (k <= bin.max) then bin.num = bin.num + v end
        end
        hist[i] = bin
    end
return hist
end

function stats.histAxis (hist)
    assert (#hist > 0, "histAxis: histogram must have values")
    local xAxis = {}
    local yAxis = {}
    for i = 1, #hist do
        xAxis[i] = hist[i].min
        yAxis[i] = hist[i].num
    end
return xAxis, yAxis
end



--[[
    Calculate the Gini Index of table of numeric values
]]
function stats.gini(t)
    assert (#t > 1, "giniIndex: table with no values")
    table.sort (t)
    local top = 0
    local bot = 0
    local num = #t
    for i = 1, num do
        top = top + (num + 1 - i) * t[i]
        bot = bot + t [i]
    end
    local g = ((num + 1 - 2*(top/bot))/num)
return g
end

return stats